@@ -7,7 +7,7 @@
  *  See also http://www.freebsd.org/copyright/freebsd-license.html
  *  --------
  *
- *  Copyright (C) 2009-2015 Erwin Waterlander
+ *  Copyright (C) 2009-2016 Erwin Waterlander
  *  Copyright (C) 1998 Christian Wurll
  *  Copyright (C) 1998 Bernd Johannes Wuebben
  *  Copyright (C) 1994-1995 Benjamin Lin.
@@ -65,6 +65,9 @@
 
 #include "common.h"
 #include "dos2unix.h"
+# if (defined(_WIN32) && !defined(__CYGWIN__))
+#include <windows.h>
+#endif
 #ifdef D2U_UNICODE
 #if !defined(__MSDOS__) && !defined(_WIN32) && !defined(__OS2__)  /* Unix, Cygwin */
 # include <langinfo.h>
@@ -73,59 +76,100 @@
 
 void PrintLicense(void)
 {
-  printf("%s", _("\
-Copyright (C) 2009-2015 Erwin Waterlander\n\
+  D2U_ANSI_FPRINTF(stdout,_("\
+Copyright (C) 2009-%d Erwin Waterlander\n\
 Copyright (C) 1998      Christian Wurll (Version 3.1)\n\
 Copyright (C) 1998      Bernd Johannes Wuebben (Version 3.0)\n\
 Copyright (C) 1994-1995 Benjamin Lin\n\
-All rights reserved.\n\n"));
+All rights reserved.\n\n"),2016);
   PrintBSDLicense();
 }
 
 #ifdef D2U_UNICODE
-void StripDelimiterW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, wint_t CurChar, unsigned int *converted)
+wint_t StripDelimiterW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, wint_t CurChar, unsigned int *converted, const char *progname)
 {
   wint_t TempNextChar;
   /* CurChar is always CR (x0d) */
   /* In normal dos2unix mode put nothing (skip CR). */
   /* Don't modify Mac files when in dos2unix mode. */
   if ( (TempNextChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
-    d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype);  /* put back peek char */
+    if (d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype) == WEOF) {  /* put back peek char */
+        d2u_getc_error(ipFlag,progname);
+        return WEOF;
+    }
     if ( TempNextChar != 0x0a ) {
-      d2u_putwc( CurChar, ipOutF, ipFlag);  /* Mac line, put back CR */
+      if (d2u_putwc(CurChar, ipOutF, ipFlag, progname) == WEOF) {  /* Mac line, put CR */
+          d2u_putwc_error(ipFlag,progname);
+          return WEOF;
+      }
     } else {
       (*converted)++;
       if (ipFlag->NewLine) {  /* add additional LF? */
-        d2u_putwc(0x0a, ipOutF, ipFlag);
+        if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) {
+            d2u_putwc_error(ipFlag,progname);
+            return WEOF;
+        }
       }
     }
+  } else {
+    if (ferror(ipInF)) {
+        d2u_getc_error(ipFlag,progname);
+        return WEOF;
+    }
+    if ( CurChar == 0x0d ) {  /* EOF: last Mac line delimiter (CR)? */
+        if (d2u_putwc(CurChar, ipOutF, ipFlag, progname) == WEOF) {
+            d2u_putwc_error(ipFlag,progname);
+            return WEOF;
+        }
+    }
   }
-  else if ( CurChar == 0x0d ) {  /* EOF: last Mac line delimiter (CR)? */
-    d2u_putwc( CurChar, ipOutF, ipFlag);
-  }
+  return CurChar;
 }
 #endif
 
-void StripDelimiter(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, int CurChar, unsigned int *converted)
+/* CUR        NEXT
+   0xd(CR)    0xa(LF)  => put LF if option -l was used
+   0xd(CR)  ! 0xa(LF)  => put CR
+   0xd(CR)    EOF      => put CR
+ */
+int StripDelimiter(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, int CurChar, unsigned int *converted, const char *progname)
 {
   int TempNextChar;
   /* CurChar is always CR (x0d) */
   /* In normal dos2unix mode put nothing (skip CR). */
   /* Don't modify Mac files when in dos2unix mode. */
   if ( (TempNextChar = fgetc(ipInF)) != EOF) {
-    ungetc( TempNextChar, ipInF );  /* put back peek char */
+    if (ungetc( TempNextChar, ipInF ) == EOF) { /* put back peek char */
+        d2u_getc_error(ipFlag,progname);
+        return EOF;
+    }
     if ( TempNextChar != '\x0a' ) {
-      fputc( CurChar, ipOutF );  /* Mac line, put back CR */
+      if (fputc( CurChar, ipOutF ) == EOF) { /* Mac line, put CR */
+          d2u_putc_error(ipFlag,progname);
+          return EOF;
+      }
     } else {
       (*converted)++;
       if (ipFlag->NewLine) {  /* add additional LF? */
-        fputc('\x0a', ipOutF);
+        if (fputc('\x0a', ipOutF) == EOF) {
+            d2u_putc_error(ipFlag,progname);
+            return EOF;
+        }
       }
     }
+  } else {
+    if (ferror(ipInF)) {
+        d2u_getc_error(ipFlag,progname);
+        return EOF;
+    }
+    if ( CurChar == '\x0d' ) {  /* EOF: last Mac line delimiter (CR)? */
+        if (fputc( CurChar, ipOutF ) == EOF) {
+            d2u_putc_error(ipFlag,progname);
+            return EOF;
+        }
+    }
   }
-  else if ( CurChar == '\x0d' ) {  /* EOF: last Mac line delimiter (CR)? */
-    fputc( CurChar, ipOutF );
-  }
+  return CurChar;
 }
 
 /* converts stream ipInF to UNIX format text and write to stream ipOutF
@@ -140,7 +184,6 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
     wint_t TempNextChar;
     unsigned int line_nr = 1;
     unsigned int converted = 0;
-    char *errstr;
 
     ipFlag->status = 0;
 
@@ -165,30 +208,30 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
               if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
-              fprintf(stderr, "%s: ", progname);
-              fprintf(stderr, _("Binary symbol 0x00%02X found at line %u\n"),TempChar, line_nr);
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x00%02X found at line %u\n"),TempChar, line_nr);
             }
             break;
           }
           if (TempChar != 0x0d) {
             if (TempChar == 0x0a) /* Count all DOS and Unix line breaks */
               ++line_nr;
-            if (d2u_putwc(TempChar, ipOutF, ipFlag) == WEOF) {
+            if (d2u_putwc(TempChar, ipOutF, ipFlag, progname) == WEOF) {
               RetVal = -1;
-              if (ipFlag->verbose) {
-                if (!(ipFlag->status & UNICODE_CONVERSION_ERROR)) {
-                  ipFlag->error = errno;
-                  errstr = strerror(errno);
-                  fprintf(stderr, "%s: ", progname);
-                  fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-                }
-              }
+              d2u_putwc_error(ipFlag,progname);
               break;
             }
           } else {
-            StripDelimiterW( ipInF, ipOutF, ipFlag, TempChar, &converted);
+            if (StripDelimiterW( ipInF, ipOutF, ipFlag, TempChar, &converted, progname) == WEOF) {
+              RetVal = -1;
+              break;
+            }
           }
         }
+        if ((TempChar == WEOF) && ferror(ipInF)) {
+          RetVal = -1;
+          d2u_getc_error(ipFlag,progname);
+        }
         break;
       case FROMTO_MAC2UNIX: /* mac2unix */
         while ((TempChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
@@ -202,68 +245,72 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
               if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
-              fprintf(stderr, "%s: ", progname);
-              fprintf(stderr, _("Binary symbol 0x00%02X found at line %u\n"), TempChar, line_nr);
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x00%02X found at line %u\n"), TempChar, line_nr);
             }
             break;
           }
           if ((TempChar != 0x0d)) {
               if (TempChar == 0x0a) /* Count all DOS and Unix line breaks */
                 ++line_nr;
-              if(d2u_putwc(TempChar, ipOutF, ipFlag) == WEOF) {
+              if(d2u_putwc(TempChar, ipOutF, ipFlag, progname) == WEOF) {
                 RetVal = -1;
-                if (ipFlag->verbose) {
-                  if (!(ipFlag->status & UNICODE_CONVERSION_ERROR)) {
-                    ipFlag->error = errno;
-                    errstr = strerror(errno);
-                    fprintf(stderr, "%s: ", progname);
-                    fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-                  }
-                }
+                d2u_putwc_error(ipFlag,progname);
                 break;
               }
             }
           else{
             /* TempChar is a CR */
             if ( (TempNextChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
-              d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype);  /* put back peek char */
+              if (d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype) == WEOF) {  /* put back peek char */
+                d2u_getc_error(ipFlag,progname);
+                RetVal = -1;
+                break;
+              }
               /* Don't touch this delimiter if it's a CR,LF pair. */
               if ( TempNextChar == 0x0a ) {
-                d2u_putwc(0x0d, ipOutF, ipFlag); /* put CR, part of DOS CR-LF */
+                if (d2u_putwc(0x0d, ipOutF, ipFlag, progname) == WEOF) { /* put CR, part of DOS CR-LF */
+                  d2u_putwc_error(ipFlag,progname);
+                  RetVal = -1;
+                  break;
+                }
                 continue;
               }
             }
-            if (d2u_putwc(0x0a, ipOutF, ipFlag) == WEOF) { /* MAC line end (CR). Put LF */
+            if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) { /* MAC line end (CR). Put LF */
                 RetVal = -1;
-                if (ipFlag->verbose) {
-                  if (!(ipFlag->status & UNICODE_CONVERSION_ERROR)) {
-                    ipFlag->error = errno;
-                    errstr = strerror(errno);
-                    fprintf(stderr, "%s: ", progname);
-                    fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-                  }
-                }
+                d2u_putwc_error(ipFlag,progname);
                 break;
               }
             converted++;
             line_nr++; /* Count all Mac line breaks */
             if (ipFlag->NewLine) {  /* add additional LF? */
-              d2u_putwc(0x0a, ipOutF, ipFlag);
+              if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) {
+                RetVal = -1;
+                d2u_putwc_error(ipFlag,progname);
+                break;
+              }
             }
           }
         }
+        if ((TempChar == WEOF) && ferror(ipInF)) {
+          RetVal = -1;
+          d2u_getc_error(ipFlag,progname);
+        }
         break;
       default: /* unknown FromToMode */
       ;
 #if DEBUG
-      fprintf(stderr, "%s: ", progname);
-      fprintf(stderr, _("program error, invalid conversion mode %d\n"),ipFlag->FromToMode);
+      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+      D2U_UTF8_FPRINTF(stderr, _("program error, invalid conversion mode %d\n"),ipFlag->FromToMode);
       exit(1);
 #endif
     }
-    if (ipFlag->verbose > 1) {
-      fprintf(stderr, "%s: ", progname);
-      fprintf(stderr, _("Converted %u out of %u line breaks.\n"), converted, line_nr -1);
+    if (ipFlag->status & UNICODE_CONVERSION_ERROR)
+        ipFlag->line_nr = line_nr;
+    if ((RetVal == 0) && (ipFlag->verbose > 1)) {
+      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+      D2U_UTF8_FPRINTF(stderr, _("Converted %u out of %u line breaks.\n"), converted, line_nr -1);
     }
     return RetVal;
 }
@@ -281,7 +328,6 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
     int *ConvTable;
     unsigned int line_nr = 1;
     unsigned int converted = 0;
-    char *errstr;
 
     ipFlag->status = 0;
 
@@ -321,8 +367,8 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
       ConvTable = D2UAsciiTable;
 
     if ((ipFlag->ConvMode > CONVMODE_7BIT) && (ipFlag->verbose)) { /* not ascii or 7bit */
-       fprintf(stderr, "%s: ", progname);
-       fprintf(stderr, _("using code page %d.\n"), ipFlag->ConvMode);
+       D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+       D2U_UTF8_FPRINTF(stderr, _("using code page %d.\n"), ipFlag->ConvMode);
     }
 
     /* CR-LF -> LF */
@@ -345,8 +391,8 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
               if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
-              fprintf(stderr, "%s: ", progname);
-              fprintf(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
             }
             break;
           }
@@ -355,18 +401,20 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
               ++line_nr;
             if (fputc(ConvTable[TempChar], ipOutF) == EOF) {
               RetVal = -1;
-              if (ipFlag->verbose) {
-                ipFlag->error = errno;
-                errstr = strerror(errno);
-                fprintf(stderr, "%s: ", progname);
-                fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-              }
+              d2u_putc_error(ipFlag,progname);
               break;
             }
           } else {
-            StripDelimiter( ipInF, ipOutF, ipFlag, TempChar, &converted);
+            if (StripDelimiter( ipInF, ipOutF, ipFlag, TempChar, &converted, progname) == EOF) {
+              RetVal = -1;
+              break;
+            }
           }
         }
+        if ((TempChar == EOF) && ferror(ipInF)) {
+          RetVal = -1;
+          d2u_getc_error(ipFlag,progname);
+        }
         break;
       case FROMTO_MAC2UNIX: /* mac2unix */
         while ((TempChar = fgetc(ipInF)) != EOF) {
@@ -380,8 +428,8 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
               if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
-              fprintf(stderr, "%s: ", progname);
-              fprintf(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
             }
             break;
           }
@@ -390,54 +438,60 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
                 ++line_nr;
               if(fputc(ConvTable[TempChar], ipOutF) == EOF) {
                 RetVal = -1;
-                if (ipFlag->verbose) {
-                  ipFlag->error = errno;
-                  errstr = strerror(errno);
-                  fprintf(stderr, "%s: ", progname);
-                  fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-                }
+                d2u_putc_error(ipFlag,progname);
                 break;
               }
             }
           else{
             /* TempChar is a CR */
             if ( (TempNextChar = fgetc(ipInF)) != EOF) {
-              ungetc( TempNextChar, ipInF );  /* put back peek char */
+              if (ungetc( TempNextChar, ipInF ) == EOF) {  /* put back peek char */
+                d2u_getc_error(ipFlag,progname);
+                RetVal = -1;
+                break;
+              }
               /* Don't touch this delimiter if it's a CR,LF pair. */
               if ( TempNextChar == '\x0a' ) {
-                fputc('\x0d', ipOutF); /* put CR, part of DOS CR-LF */
+                if (fputc('\x0d', ipOutF) == EOF) { /* put CR, part of DOS CR-LF */
+                  RetVal = -1;
+                  d2u_putc_error(ipFlag,progname);
+                  break;
+                }
                 continue;
               }
             }
             if (fputc('\x0a', ipOutF) == EOF) { /* MAC line end (CR). Put LF */
                 RetVal = -1;
-                if (ipFlag->verbose) {
-                  ipFlag->error = errno;
-                  errstr = strerror(errno);
-                  fprintf(stderr, "%s: ", progname);
-                  fprintf(stderr, _("can not write to output file: %s\n"), errstr);
-                }
+                d2u_putc_error(ipFlag,progname);
                 break;
               }
             converted++;
             line_nr++; /* Count all Mac line breaks */
             if (ipFlag->NewLine) {  /* add additional LF? */
-              fputc('\x0a', ipOutF);
+              if (fputc('\x0a', ipOutF) == EOF) {
+                RetVal = -1;
+                d2u_putc_error(ipFlag,progname);
+                break;
+              }
             }
           }
         }
+        if ((TempChar == EOF) && ferror(ipInF)) {
+          RetVal = -1;
+          d2u_getc_error(ipFlag,progname);
+        }
         break;
       default: /* unknown FromToMode */
       ;
 #if DEBUG
-      fprintf(stderr, "%s: ", progname);
-      fprintf(stderr, _("program error, invalid conversion mode %d\n"),ipFlag->FromToMode);
+      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+      D2U_UTF8_FPRINTF(stderr, _("program error, invalid conversion mode %d\n"),ipFlag->FromToMode);
       exit(1);
 #endif
     }
-    if (ipFlag->verbose > 1) {
-      fprintf(stderr, "%s: ", progname);
-      fprintf(stderr, _("Converted %u out of %u line breaks.\n"),converted, line_nr -1);
+    if ((RetVal == 0) && (ipFlag->verbose > 1)) {
+      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+      D2U_UTF8_FPRINTF(stderr, _("Converted %u out of %u line breaks.\n"),converted, line_nr -1);
     }
     return RetVal;
 }
@@ -453,6 +507,12 @@ int main (int argc, char *argv[])
 # ifdef __MINGW64__
   int _dowildcard = -1; /* enable wildcard expansion for Win64 */
 # endif
+  int  argc_new;
+  char **argv_new;
+#ifdef D2U_UNIFILE
+  wchar_t **wargv;
+  char ***argv_glob;
+# endif
 
   progname[8] = '\0';
   strcpy(progname,"dos2unix");
@@ -465,8 +525,8 @@ int main (int argc, char *argv[])
       if (strlen(ptr) < sizeof(localedir))
          strcpy(localedir,ptr);
       else {
-         fprintf(stderr,"%s: ",progname);
-         fprintf(stderr, "%s", _("error: Value of environment variable DOS2UNIX_LOCALEDIR is too long.\n"));
+         D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
+         D2U_ANSI_FPRINTF(stderr, "%s", _("error: Value of environment variable DOS2UNIX_LOCALEDIR is too long.\n"));
          strcpy(localedir,LOCALEDIR);
       }
    }
@@ -474,8 +534,16 @@ int main (int argc, char *argv[])
 
 #if defined(ENABLE_NLS) || (defined(D2U_UNICODE) && !defined(__MSDOS__) && !defined(_WIN32) && !defined(__OS2__))
 /* setlocale() is also needed for nl_langinfo() */
+#if (defined(_WIN32) && !defined(__CYGWIN__))
+/* When the locale is set to "" on Windows all East-Asian multi-byte ANSI encoded text is printed
+   wrongly when you use standard printf(). Also UTF-8 code is printed wrongly. See also test/setlocale.c.
+   When we set the locale to "C" gettext still translates the messages on Windows. On Unix this would disable
+   gettext. */
+   setlocale (LC_ALL, "C");
+#else
    setlocale (LC_ALL, "");
 #endif
+#endif
 
 #ifdef ENABLE_NLS
    bindtextdomain (PACKAGE, localedir);
@@ -485,6 +553,11 @@ int main (int argc, char *argv[])
 
   /* variable initialisations */
   pFlag = (CFlag*)malloc(sizeof(CFlag));
+  if (pFlag == NULL) {
+    D2U_UTF8_FPRINTF(stderr, "dos2unix:");
+    D2U_ANSI_FPRINTF(stderr, " %s\n", strerror(errno));
+    return errno;
+  }
   pFlag->FromToMode = FROMTO_DOS2UNIX;  /* default dos2unix */
   pFlag->keep_bom = 0;
 
@@ -498,10 +571,31 @@ int main (int argc, char *argv[])
     strcpy(progname,"mac2unix");
   }
 
+#ifdef D2U_UNIFILE
+  /* Get arguments in wide Unicode format in the Windows Command Prompt */
+
+  /* This does not support wildcard expansion (globbing) */
+  wargv = CommandLineToArgvW(GetCommandLineW(), &argc);
+
+  argv_glob = (char ***)malloc(sizeof(char***));
+  if (argv_glob == NULL) {
+    D2U_UTF8_FPRINTF(stderr, "%s:", progname);
+    D2U_ANSI_FPRINTF(stderr, " %s\n", strerror(errno));
+    free(pFlag);
+    return errno;
+  }
+  /* Glob the arguments and convert them to UTF-8 */
+  argc_new = glob_warg(argc, wargv, argv_glob, pFlag, progname);
+  argv_new = *argv_glob;
+#else  
+  argc_new = argc;
+  argv_new = argv;
+#endif
+
 #ifdef D2U_UNICODE
-  return parse_options(argc, argv, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix, ConvertDosToUnixW);
+  return parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix, ConvertDosToUnixW);
 #else
-  return parse_options(argc, argv, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix);
+  return parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix);
 #endif
 }
 
