@@ -28,6 +28,7 @@
 #ifndef __UNIX2DOS_H
 #define __UNIX2DOS_H
 
+#ifdef __UNIX2DOS_C
 /* ASCII mode. No conversion. */
 
 static int U2DAsciiTable[256] =
@@ -212,5 +213,11 @@ static int U2DIso1252Table[256] =
   '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7', '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef',
   '\xf0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xf7', '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xff',
 };
+#endif
+
+#ifdef D2U_UNICODE
+int ConvertUnixToDosW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progname);
+#endif
+int ConvertUnixToDos(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progname);
 
 #endif
